﻿local FS = LibStub("AceAddon-3.0"):NewAddon("FaceShooter", "AceConsole-3.0", "AceEvent-3.0")

-- defaults used if FS is started for the first time on a given char
FS.defaults = {
	profile = {
		lock = false,
		showInRaid = true,
		showInParty = true,
		showWhenSolo = true,
		maxWait = 200,
		damping = 2,
		["debug"] = {
			point = "CENTER",
			x = 0,
			y = 0,
		},
		["sps"] = {
			suggestSpS = "always",
			saveFocusSpS = true,
		},
		["ba"] = {
			suggestBA = "always",
			suggestET = "always",
			saveFocusBA = true,
			saveFocusET = true,
			useTraps = "never",
		},
		["hm"] = {
			suggestHM = "always",
			minTimeHM = 15,
		},
		["aims"] = {
			aimsBreakpoint = 1.7,
		},
		["T12"] = {
			replaceAS = true,
		},
		["T13"] = {
			forceAS = false,
		},
		["button1"] = {
			enabled = true,
			point = "CENTER",
			x = -100,
			y = 0,
			scale = 100,
			alpha = 100,
		},
		["button2"] = {
			enabled = true,
			wasEnabled = true,
			point = "CENTER",
			x = -50,
			y = 0,
			scale = 100,
			alpha = 100,
		},
	},
}

-- initializes FS - called when the addon is loaded
function FS:OnInitialize()
	-- don´t load FS unless player is at least lvl 80 and hunter
	local _, playerClass = UnitClass("player")
	local unitLevel = UnitLevel("player")
	if playerClass ~= "HUNTER" or unitLevel < 80 then
		self:Print("FaceShooter: Only Hunters of lvl 80+ supported. You should deactivate the addon on this character.")
		self:Disable()
		return false
	end

	-- read the savedvariables
	self.db = LibStub("AceDB-3.0"):New("FSDB", FS.defaults)
	
	-- DB data conversion to 1.0.8
	if self.db.profile.aims.noCaBreakpoint ~= nil and (self.db.profile.aims.noCaBreakpoint < 1 or self.db.profile.aims.noCaBreakpoint > 2.9) then
		-- Check wether the AimS settings are old and print a reminder if so
		self:Print("Aimed Shot breakpoint definition has been changed from haste percentage to casttime. Please check your Aimed Shot options and set them to appropriate values")
	end

	-- DB data conversion to 1.1.0 (Pre MoP Patch 5.0.4)
	if self.db.profile.aims ~= nil then
		-- CA no longer exists. Use noCaBreakpoint as general aimsBreakpoint if possible.
		if self.db.profile.aims.caBreakpoint ~= nil then self.db.profile.aims.caBreakpoint = nil end
		if self.db.profile.aims.caUseStyleOver ~= nil then self.db.profile.aims.caUseStyleOver = nil end
		if self.db.profile.aims.caUseStyleUnder ~= nil then self.db.profile.aims.caUseStyleUnder = nil end
		if self.db.profile.aims.noCaBreakpoint ~= nil then
			self.db.profile.aims.aimsBreakpoint = self.db.profile.aims.noCaBreakPoint
			self.db.profile.aims.noCaBreakpoint = nil
		end
		if self.db.profile.aims.noCaUseStyleOver ~= nil then self.db.profile.aims.noCaUseStyleOver = nil end
		if self.db.profile.aims.noCaUseStyleUnder ~= nil then self.db.profile.aims.noCaUseStyleUnder = nil end
	end
	if self.db.profile.lnl ~= nil then
		-- There's no need for an explicit focus dump since ES can be cast back on back now
		if self.db.profile.lnl.useKCasdump ~= nil then self.db.profile.lnl.useKCasdump = nil end
	end

	-- ACE3 style options table
	FS.options = {
		type = "group",
		name = "FaceShooter",
		handler = FS,
		args = {
			debug = {
				type = "toggle",
				name = "debug",
				get = "IsDebug",
				set = "ToggleDebug",
				hidden = true,
			},
			log = {
				type = "toggle",
				name = "log",
				get = "IsLog",
				set = "ToggleLog",
				hidden = true,
			},
			lock = {
				type = "toggle",
				name = "Lock frames",
				get = "IsLock",
				set = "ToggleLock",
				width = "full",
				order = 1,
			},
			showInRaid = {
				type = "toggle",
				name = "Show in Raid",
				get = "IsShowInRaid",
				set = "ToggleShowInRaid",
				order = 2,
			},
			showInPary = {
				type = "toggle",
				name = "Show in party",
				get = "IsShowInParty",
				set = "ToggleShowInParty",
				order = 3,
			},
			showWhenSolo = {
				type = "toggle",
				name = "Show when solo",
				get = "IsShowWhenSolo",
				set = "ToggleShowWhenSolo",
				order = 4,
			},
			maxWait = {
				type = "range",
				name = "Max. time to wait for signature shots (in ms)",
				min = 0,
				max = 1000,
				step = 100,
				width = "double",
				get = "GetMaxWait",
				set = "SetMaxWait",
				width = "full",
				order = 5,
			},
			sps = {
				type = "group",
				name = "Serpent Sting Options",
				order = 6,
				args = {
					suggestSpS = {
						type = "select",
						name = "Suggest Serpent Sting",
						values = {
							never = "Never",
							level = "Bosslevel",
							always = "Always",
						},
						get = "GetSuggestSpS",
						set = "SetSuggestSpS",
						order = 1,
					},
					saveFocusSpS = {
						type = "toggle",
						name = "Save focus for sig. shot when suggesting Serpent Sting",
						get = "IsSaveFocusSpS",
						set = "ToggleSaveFocusSpS",
						width = "full",
						order = 2,
					},
				},
			},
			ba = {
				type = "group",
				name = "Black Arrow Options",
				order = 7,
				args = {
					suggestBA = {
						type = "select",
						name = "Suggest Black Arrow",
						values = {
							never = "Never",
							level = "Bosslevel",
							always = "Always",
						},
						get = "GetSuggestBA",
						set = "SetSuggestBA",
						order = 1,
					},
					suggestET = {
						type = "select",
						name = "Suggest Explosive Trap",
						values = {
							never = "Never",
							level = "Bosslevel",
							always = "Always",
						},
						get = "GetSuggestET",
						set = "SetSuggestET",
						order = 2,
					},
					useTraps = {
						type = "select",
						name = "Use Explosive Traps in rotation",
						values = {
							never = "Never",
							asSV = "As Survival",
							always = "Always",
						},
						get = "GetUseTraps",
						set = "SetUseTraps",
						order = 3,
					},
					saveFocusBA = {
						type = "toggle",
						name = "Save focus for sig. shot when suggesting Black Arrow",
						get = "IsSaveFocusBA",
						set = "ToggleSaveFocusBA",
						width = "full",
						order = 4
					},
					saveFocusET = {
						type = "toggle",
						name = "Save focus for sig. shot when suggesting Explosive Trap",
						get = "IsSaveFocusET",
						set = "ToggleSaveFocusET",
						width = "full",
						order = 5,
					},
				},
			},
			hm = {
				type = "group",
				name = "Hunter's Mark Options",
				order = 8,
				args = {
					suggestHM = {
						type = "select",
						name = "suggest Hunter's Mark",
						values = {
							never = "Never",
							ooc = "Ooc",
							level = "Bosslevel",
							oocORlevel = "Ooc or bosslevel",
							oocANDlevel = "Ooc and bosslevel",
							always = "Always",
						},
						get = "GetSuggestHM",
						set = "SetSuggestHM",
						order = 1,
					},
					minTimeHM = {
						type = "range",
						name = "Min. time before suggesting to reapply Hunter's Mark",
						min = 0,
						max = 60,
						step = 1,
						width = "double",
						get = "GetMinTimeHM",
						set = "SetMinTimeHM",
						order = 2,
					},
				},
			},
			aims = {
				type = "group",
				name = "Aimed Shot Options",
				order = 9,
				args = {
					aimsBreakpoint = {
						type = "range",
						name = "Haste breakpoint at casttime",
						width = "full",
						min = 1,
						max = 2.9,
						step = 0.1,
						get = "GetAimBreakpoint",
						set = "SetAimBreakpoint",
						order = 1,
					},
				},
			},
			T12 = {
				type = "group",
				name = "T12 proc Options",
				order = 10,
				args = {
					replaceAS = {
						type = "toggle",
						name = "Replace Arcane shot with Kill Command on proc",
						get = "IsReplaceAS",
						set = "ToggleReplaceAS",
						width = "full",
						order = 1,
					},
				},
			},
			T13 = {
				type = "group",
				name = "T13 proc Options",
				order = 11,
				args = {
					forceAS = {
						type = "toggle",
						name = "Force using Arcane over Aimed shot when proc is off ICD",
						get = "IsForceAS",
						set = "ToggleForceAS",
						width = "full",
						order = 1,
					},
				},
			},
		},
	}

	-- extend the options table to the necessary amount of shots
	for i = 1, 2 do
		FS.options.args["button"..i] = {
			type = "group",
			name = "Shot recom. Frame "..i,
			args = {
				enabled = {
					type = "toggle",
					name = "Enable frame",
					get = "IsEnabled",
					set = "ToggleEnabled",
					width = "full",
					order = 1,
				},
				scale = {
					type = "range",
					name = "Frame scale",
					min = 50,
					max = 200,
					bigStep = 10,
					get = "GetScale",
					set = "SetScale",
					order = 2,
				},
				alpha = {
					type = "range",
					name = "Frame alpha",
					min = 10,
					max = 100,
					bigStep = 5,
					get = "GetAlpha",
					set = "SetAlpha",
					order = 3,
				},
			},
		}
		if i == 1 then FS.options.args.button1.args.enabled = nil
		else
			FS.options.args["button"..i].args.scale.disabled = not self.db.profile["button"..i]["enabled"]
			FS.options.args["button"..i].args.alpha.disabled = not self.db.profile["button"..i]["enabled"]
		end		
	end

	-- disable options as required
	self:CheckForDisabled()
	
	-- setup the options table based on the options table
	LibStub("AceConfig-3.0"):RegisterOptionsTable("FaceShooter", FS.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("FaceShooter", "FaceShooter")
	self:RegisterChatCommand("fs", "ChatCommand")
	self:RegisterChatCommand("faceshooter", "ChatCommand")
end

-- starts FS up - called when the addon is enabled
function FS:OnEnable()
	-- some FS global vars
	self.version = "1.1.2"
	self.state = "idle"
	self.tslu = 0
	self.ooc = not UnitAffectingCombat("player")
	self.level = UnitLevel("player")
	self.guid = UnitGUID("player")
	self.gcd = 0
	self.avrgLatency = 0
	self.debug = false
	self.spellNum = {}
	self.spellName = {}
	self.spellSentAt = {}
	self.lastCastSentAt = 0
	self.spellStartedAt = {}
	self.lastCastStartedAt = 0
	self.lastCastSucceededAt = 0
	self.lastT13procAt = 0
	self.blockDB = 0
	self.spellEndedAt = {}
	self.cost = {}
	self.freeTables = {}
	
	-- the polling intervals for the different states
	self.intervals = {
		["idle"] = 1,
		["castSent"] = 0.5,
		["castStarted"] = 0.1,
		["succeeded"] = 0.1,
		["latency"] = 0.5,
		["waiting"] = 0.5,
	}

	-- spellIDs of used spells and talents
	self.spellID = {
		-- Shots and spells
		["SS"] = 56641, -- Steady Shot
		["CoS"] = 77767, -- Cobra Shot
		["SpS"] = 1978, -- Serpent Sting
		["CS"] = 53209, -- Chimera Shot
		["BA"] = 3674, -- Black Arrow
		["ES"] = 53301, -- Explosive Shot
		["AimS"] = 82928, -- Aimed Shot
		["AS"] = 3044, -- Arcane Shot
		["KS"] = 53351, -- Kill Shot
		["KC"] = 34026, -- Kill Command
		["ET"] = 82939, -- Explosive Trap
		["MS"] = 2643, -- Multi-Shot
		["TQ"] = 19801, -- Tranquilizing Shot
		["DB"] = 120679, -- Dire Beast
		["AMoC"] = 131894, -- A Murder of Crows
		["BS"] = 130392, -- Blink Strike
		["LR"] = 120697, -- Lynx Rush
		["GT"] = 117050, -- Glaive Toss
		["PS"] = 109259, -- Power Shot
		["BRG"] = 120360, -- Barrage
		-- Off GCD spells
		["AutoS"] = 75, -- Auto Shot
		["TL"] = 77769, -- Trap Launcher
		["RF"] = 3045, -- Rapid Fire
		-- Buffs
		["LnL"] = 56453, -- Lock and Load
		["TotH"] = 34720, -- Thrill of the Hunt
		["MMM"] = 82926, -- Master Marksman
		["SF"] = 53220, -- Steady Focus
		["TBW"] = 34692, -- The Beast Within
		["BrA"] = 99060, -- Burning Adrenaline (4T12 proc)
		["CHR"] = 105919, -- Chrono Hunter (4T13 proc)
		-- Debuffs
		["HM"] = 1130, -- Hunter's Mark
	}

	-- glyphIDs of used glyphs
	self.glyphID = {
		["CS"] = 63065, -- Glyph of Chimera Shot
		["KS"] = 63067, -- Glyph of Kill Shot
		["KC"] = 56842, -- Glyph of Kill Command
		["TL"] = 56857, -- Glyph of Trap Launcher
	}

	-- itemIDs for set items that need to be considered
	self.itemID = {
		-- Wyrmstamker Battlegear (T13)
		["T13"] = {
			-- Raid Finder
			["RF_Head"] = 78793,
			["RF_Shoulder"] = 78832,
			["RF_Chest"] = 78756,
			["RF_Hands"] = 78769,
			["RF_Legs"] = 78804,
			-- Normal
			["N_Head"] = 77030,
			["N_Shoulder"] = 77032,
			["N_Chest"] = 77028,
			["N_Hands"] = 77029,
			["N_Legs"] = 77031,
			-- Heroic
			["HC_Head"] = 78698,
			["HC_Shoulder"] = 78737,
			["HC_Chest"] = 78661,
			["HC_Hands"] = 78674,
			["HC_Legs"] = 78709,
		},
		-- Lightning-Charged Battlegear (T11)
		["T11"] = {
			-- Normal
			["N_Head"] = 60303,
			["N_Shoulder"] = 60306,
			["N_Chest"] = 60304,
			["N_Hands"] = 60307,
			["N_Legs"] = 60305,
			-- Heroic
			["HC_Head"] = 65206,
			["HC_Shoulder"] = 65208,
			["HC_Chest"] = 65204,
			["HC_Hands"] = 65205,
			["HC_Legs"] = 65207,
		}
	}
	
	-- register for needed events
	self:RegisterEvent("PLAYER_ALIVE")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TALENT_UPDATE")
	self:RegisterEvent("GLYPH_ADDED")
	self:RegisterEvent("GLYPH_REMOVED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_FACTION")
	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
	
	self.frame = CreateFrame("Frame")
	self.frame:SetScript("OnUpdate", function(this, elapsed) self:OnUpdate(elapsed) end)

	-- create the base resources object
	self.resources = self:CreateResources()
	
	-- setup the button frames
	for i = 1, 2 do
		if self.db.profile["button"..i]["enabled"] then self["button"..i] = self:CreateButton("button"..i) end
	end
end

-- disables FS
function FS:OnDisable()
    -- Called when the addon is disabled
	self:UnregisterEvent("PLAYER_TALENT_UPDATE")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UNIT_FACTION")
	for name, module in FS:IterateModules() do
		module:DisableModule()
	end
end

local timeStamp = 0
-- basetime for FS that calls the scheduler at given intervals
function FS:OnUpdate(elapsed)
	self.tslu = self.tslu + elapsed
	if self.tslu > self.intervals[self.state] then
		self:Scheduler()
	end
end

-- schedules the different tasks of FS
function FS:Scheduler()
	local now, i = GetTime()
	-- set time since last update to 0 (it´s beeing updated now!)
	self.tslu = 0
	
	-- update the debugframe
	if self.debug then
		self.debugFrm:GetDebugStr()
		self.debugFrm:WriteText()
	end
	
	-- set to idle state when mounting
	if self:IsMounted() and not self.isMounted then
		self.isMounted = true
		self:LooseTarget()
	-- set to active state when unmounting
	elseif not self:IsMounted() and self.isMounted then
		self.isMounted = false
		if self:IsLegitTarget() then self:AcquireTarget() end
	end
	
	-- break if in idle state
	if self.state == "idle" then
		return nil
	end
		
	-- update the debugframe
	if self.debug then
		if self.lastState ~= self.state then
			print("["..self:Round(now, 3).."] In "..self.state.." state now")
			self.lastState = self.state
		end
		self.debugFrm:GetDebugStr()
		self.debugFrm:WriteText()
		if self.log then self:WriteLogEntry("Before") end
	end
	
	-- check for interrupts
	self:CheckForInterrups()
	
	-- check if gcd is up and set state accordingly
	if self.state == "succeeded" and self.gcd <= now then
		self.state = "waiting"
		if self.spellName[1] then
			table.insert(self.spellName, 1, nil)
			table.remove(self.spellName, 5)
			table.insert(self.spellSentAt, 1, nil)
			table.remove(self.spellSentAt, 5)
			table.insert(self.spellStartedAt, 1, nil)
			table.remove(self.spellStartedAt, 5)
			table.insert(self.spellEndedAt, 1, nil)
			table.remove(self.spellEndedAt, 5)
		end
	end

	-- in castSent state get current resources, calculate resulting resources if that spell is cast
	if self.state == "castSent" then
		self.resources:GetResources()
		self.resources:CastSpell("fs")
		self.button1.resources:SetResources(self.resources)
		self.button1.resources:FocusGains()
		self.button1:DrawCDF(self:GetGCD())
	-- in succeeded and castStarted state most resources should be accurate according to the api
	elseif self.state == "succeeded" or self.state == "castStarted" then
		self.resources:GetResources()
		self.resources:CastSpell("fs")
		self.button1.resources:SetResources(self.resources)
		self.button1.resources:FocusGains()
	-- in waiting state we can use current resources unmodified
	elseif self.state == "waiting" then
		self.resources:GetResources()
		self.button1.resources:SetResources(self.resources)
		self.button1.resources:FocusGains()
	end
	
	-- pick the most suitable spell with those resources
	if self.state ~= "latency" then
		for i = 1, 2 do
			local thisButton, nextButton = "button"..i, "button"..(i + 1)
			self[thisButton].resources:PickSpell(thisButton)
			if i < 2 and self.db.profile[nextButton].enabled then
				self[nextButton].resources:SetResources(self[thisButton].resources)
				self[nextButton].resources:CastSpell(nextButton)
				self[nextButton].resources:FocusGains()
			else
				break
			end
		end
	end
	
	-- update the debugframe
	if self.debug then
		if self.lastState ~= self.state then
			print("["..self:Round(now, 3).."] In "..self.state.." state now")
			self.lastState = self.state
		end
		self.debugFrm:GetDebugStr()
		self.debugFrm:WriteText()
		if self.log then self:WriteLogEntry("After") end
	end
	
	-- after cast has been sent and scheduler is done change to latency state
	if self.state == "castSent" then self.state = "latency" end	
	
end
