local FS = LibStub("AceAddon-3.0"):GetAddon("FaceShooter")

-- setup slash commmands
function FS:ChatCommand(input)
    if not input or input:trim() == "" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
    else
		LibStub("AceConfigCmd-3.0").HandleCommand(FS, "fs", "FaceShooter", input)
    end
end

function FS:IsDebug(info)
	return self.debug
end

function FS:ToggleDebug(info, value)
	if value then
		self:Print("Debug mode activated.")
		if self.debugFrm then self.debugFrm.btn:Show()
		else self.debugFrm = self:CreateDebugFrame()
		end
	else
		self:Print("Debug mode deactivated.")
		self.debugFrm.btn:Hide()
	end
	self.debug = value
end

function FS:IsLog(info)
	return self.log
end

function FS:ToggleLog(info, value)
	if value then
		self:Print("Log mode activated.")
		self.db.profile.log = {}
	else
		self:Print("Log mode deactivated.")
		self.db.profile.log = nil
	end
	self.log = value
end

function FS:IsLock(info)
	return self.db.profile.lock
end

function FS:ToggleLock(info, value)
	local i
	if value then
		for i = 1, 2 do
			if self.db.profile["button"..i]["enabled"] then
				self["button"..i]["mov"]:EnableMouse(false)
				self["button"..i]["mov"]:SetAlpha(0)
			end
		end
		for name, module in FS:IterateModules() do
			if module:IsEnabled() and module.Lock then module:Lock() end
		end
		self:Print("Frames locked.")
	else
		for i = 1, 2 do
			if self.db.profile["button"..i]["enabled"] then
				self["button"..i]["mov"]:EnableMouse(true)
				self["button"..i]["mov"]:SetAlpha(1)
			end
		end
		for name, module in FS:IterateModules() do
			if module:IsEnabled() and module.UnLock then module:UnLock() end
		end
		self:Print("Frames unlocked.")
	end
	self.db.profile.lock = value
end

function FS:IsShowInRaid(info)
	return self.db.profile.showInRaid
end

function FS:ToggleShowInRaid(info, value)
	local usedToBeLegitTarget = self:IsLegitTarget()
	self.db.profile.showInRaid = value
	if value then self:Print("Frames are shown when in raid now.")
	else self:Print("Frames are no longer shown when in raid now.")
	end
	if self:IsLegitTarget() and not usedToBeLegitTarget then self:AcquireTarget()
	elseif not self:IsLegitTarget() and usedToBeLegitTarget then self:LooseTarget()
	end
end

function FS:IsShowInParty(info)
	return self.db.profile.showInParty
end

function FS:ToggleShowInParty(info, value)
	local usedToBeLegitTarget = self:IsLegitTarget()
	self.db.profile.showInParty = value
	if value then self:Print("Frames are shown when in party now.")
	else self:Print("Frames are no longer shown when in party now.")
	end
	if self:IsLegitTarget() and not usedToBeLegitTarget then self:AcquireTarget()
	elseif not self:IsLegitTarget() and usedToBeLegitTarget then self:LooseTarget()
	end
end

function FS:IsShowWhenSolo(info)
	return self.db.profile.showWhenSolo
end

function FS:ToggleShowWhenSolo(info, value)
	local usedToBeLegitTarget = self:IsLegitTarget()
	self.db.profile.showWhenSolo = value
	if value then self:Print("Frames are shown when solo now.")
	else self:Print("Frames are no longer shown when solo now.")
	end
	if self:IsLegitTarget() and not usedToBeLegitTarget then self:AcquireTarget()
	elseif not self:IsLegitTarget() and usedToBeLegitTarget then self:LooseTarget()
	end
end

function FS:GetMaxWait(info)
	return self.db.profile.maxWait
end

function FS:SetMaxWait(info, value)
	self.db.profile.maxWait = value
	if info.uiType == "cmd" then self:Print("Maximum time to wait for signature shot to cooldown set to "..value.."ms.") end
end

function FS:GetUseTraps(info)
	return self.db.profile.ba.useTraps
end

function FS:SetUseTraps(info, value)
	if info.uiType == "cmd" and value ~= self.db.profile.ba.useTraps then
		if value == "never" then self:Print("Not using Explosive Trap in rotations.")
		elseif value == "asSV" then self:Print("Only using Explosive Trap as replacement for Black Arrow.")
		else self:Print("Using Explosive Trap in all rotations.")
		end
	end
	self.db.profile.ba.useTraps = value
	self:CheckForDisabled()
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.CheckForDisabled then module:CheckForDisabled() end
	end
end

function FS:GetMinTimeHM(info)
	return self.db.profile.hm.minTimeHM
end

function FS:SetMinTimeHM(info, value)
	self.db.profile.hm.minTimeHM = value
	if info.uiType == "cmd" then self:Print("Minimum time before suggesting to reapply Hunter's Mark set to "..value.."s") end
end

function FS:GetSuggestHM(info)
	return self.db.profile.hm.suggestHM
end

function FS:SetSuggestHM(info, value)
	if info.uiType == "cmd" and value ~= self.db.profile.hm.suggestHM then
		if value == "never" then self:Print("Never suggesting Hunter's Mark in rotations.")
		elseif value == "ooc" then self:Print("Only suggest Hunter's Mark when out of combat.")
		elseif value == "level" then self:Print("Only suggest Hunter's Mark when target has bosslevel.")
		elseif value == "oocORlevel" then self:Print("Only suggest Hunter's Mark when out of combat or target has bosslevel.")
		elseif value == "oocANDlevel" then self:Print("Only suggest Hunter's Mark when out of combat and target has bosslevel.")
		else self:Print("Always suggest Hunter's Mark.")
		end
	end
	self.db.profile.hm.suggestHM = value
	self:CheckForDisabled()
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.CheckForDisabled then module:CheckForDisabled() end
	end
end

function FS:GetSuggestSpS(info)
	return self.db.profile.sps.suggestSpS
end

function FS:SetSuggestSpS(info, value)
	if info.uiType == "cmd" and value ~= self.db.profile.sps.suggestSpS then
		if value == "never" then self:Print("Never suggesting Serpent Sting in rotations.")
		elseif value == "level" then self:Print("Only suggest Serpent Sting when target has bosslevel.")
		else self:Print("Always suggest Serpent Sting.")
		end
	end
	self.db.profile.sps.suggestSpS = value
	self:CheckForDisabled()
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.CheckForDisabled then module:CheckForDisabled() end
	end
end

function FS:GetAimBreakpoint(info)
	return self.db.profile.aims.aimsBreakpoint
end

function FS:SetAimBreakpoint(info, value)
	self.db.profile.aims.aimBreakpoint = value
	if info.uiType == "cmd" then self:Print("Casttime breakpoint for using Aimed Shot set to "..value.."s") end
end

function FS:GetSuggestBA(info)
	return self.db.profile.ba.suggestBA
end

function FS:SetSuggestBA(info, value)
	if info.uiType == "cmd" and value ~= self.db.profile.ba.suggestBA then
		if value == "never" then self:Print("Never suggesting Black Arrow in rotations.")
		elseif value == "level" then self:Print("Only suggest Black Arrow when target has bosslevel.")
		else self:Print("Always suggest Black Arrow.")
		end
	end
	self.db.profile.ba.suggestBA = value
	self:CheckForDisabled()
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.CheckForDisabled then module:CheckForDisabled() end
	end
end

function FS:GetSuggestET(info)
	return self.db.profile.ba.suggestET
end

function FS:SetSuggestET(info, value)
	if info.uiType == "cmd" and value ~= self.db.profile.ba.suggestET then
		if value == "never" then self:Print("Never suggesting Explosive Trap in rotations.")
		elseif value == "level" then self:Print("Only suggest Explosive Trap when target has bosslevel.")
		else self:Print("Always suggest Explosive Trap.")
		end
	end
	self.db.profile.ba.suggestET = value
	self:CheckForDisabled()
	for name, module in FS:IterateModules() do
		if module:IsEnabled() and module.CheckForDisabled then module:CheckForDisabled() end
	end
end

function FS:IsSaveFocusSpS(info)
	return self.db.profile.sps.saveFocusSpS
end

function FS:ToggleSaveFocusSpS(info, value)
	self.db.profile.sps.saveFocusSpS = value
	if value then self:Print("Saving focus for signature shot when suggesting Serpent Sting now.")
	else self:Print("No longer saving focus for signature shot when suggesting Serpent Sting now.")
	end
end

function FS:IsSaveFocusBA(info)
	return self.db.profile.ba.saveFocusBA
end

function FS:ToggleSaveFocusBA(info, value)
	self.db.profile.ba.saveFocusBA = value
	if value then self:Print("Saving focus for signature shot when suggesting Black Arrow now.")
	else self:Print("No longer saving focus for signature shot when suggesting Black Arrow now.")
	end
end

function FS:IsSaveFocusET(info)
	return self.db.profile.ba.saveFocusET
end

function FS:ToggleSaveFocusET(info, value)
	self.db.profile.ba.saveFocusET = value
	if value then self:Print("Saving focus for signature shot when suggesting Explosive Trap now.")
	else self:Print("No longer saving focus for signature shot when suggesting Explosive Trap now.")
	end
end

function FS:IsReplaceAS(info)
	return self.db.profile.T12.replaceAS
end

function FS:ToggleReplaceAS(info, value)
	self.db.profile.T12.replaceAS = value
	if value then self:Print("Replacing Arcane Shot with Kill Command on T12 proc now.")
	else self:Print("No longer replace Arcane shot with Kill Command on T12 proc.")
	end
end

function FS:IsForceAS(info)
	return self.db.profile.T13.forceAS
end

function FS:ToggleForceAS(info, value)
	self.db.profile.T13.forceAS = value
	if value then self:Print("Force using Arcane Shot instead of Aimed shot when proc is off ICD.")
	else self:Print("No longer force using Arcane Shot instead of Aimed shot when proc is off ICD.")
	end
end

function FS:IsEnabled(info)
	return self.db.profile[(info[#info-1])]["enabled"]
end

function FS:ToggleEnabled(info, value)
	local name = (info[#info-1])
	local i
	if value then
		if name == "button1" or name == "button2" then
			self:Print("Enabled second recommendation frame.")
			self.button2 = self.button2 or self:CreateButton("button2")
			self.db.profile.button2.enabled = true
			self.db.profile.button2.wasEnabled = true
		else
			self:Print(self.modNames[name].." module enabled.")
			self.db.profile[name]["enabled"] = true
			FS:EnableModule(self.modNames[name])
		end
	else
		if name == "button1" or name == "button2" then
			self:Print("Disabled second recommendation frame.")
			if self.button2 then self:RemoveButton("button2") end
			self.db.profile.button2.enabled = false
			self.db.profile.button2.wasEnabled = false
		else
			self:Print(self.modNames[name].." module disabled.")
			self.db.profile[name]["enabled"] = false
			FS:DisableModule(self.modNames[name])
		end
	end
end

function FS:GetScale(info)
	return self.db.profile[(info[#info-1])]["scale"]
end

function FS:SetScale(info, value)
	local name = info[#info-1]
	if name == "button1" or name == "button2" then
		self[name]["mov"]:SetScale(value/100)
		self[name]["btn"]:SetScale(value/100)
	else
		local module = FS:GetModule(self.modNames[name])
		module.mov:SetScale(value/100)
	end
	self.db.profile[(info[#info-1])]["scale"] = value
	if info.uiType == "cmd" then self:Print("Scale set to "..value.."%") end
end

function FS:GetAlpha(info)
	return self.db.profile[(info[#info-1])]["alpha"]
end

function FS:SetAlpha(info, value)
	local name = info[#info-1]
	if name == "button1" or name == "button2" then
		self[(info[#info-1])]["btn"]:SetAlpha(value/100)
	else
		local module = FS:GetModule(self.modNames[name])
		module.frm:SetAlpha(value/100)
	end
	self.db.profile[(info[#info-1])]["alpha"] = value
	if info.uiType == "cmd" then self:Print("Alpha set to "..value.."%") end
end
