﻿local FS = LibStub("AceAddon-3.0"):GetAddon("FaceShooter")
local FS_TTL = FS:NewModule("TimeToLife")
local now = GetTime

--------------
--- FS_TTL ---
--------------

-- TTL module specific defaults
FS.defaults.profile.ttl = {
	enabled = true,
	shown = true,
	updInterval = 0.5,
	evalSamples = 9,
	numSamples = 20,
	assumedDps = 10000,
	point = "CENTER",
	x = -50,
	y = -100,
	scale = 100,
	alpha = 100,
}
FS.defaults.profile.sps.minTTLSpS = 15
FS.defaults.profile.ba.minTTLBA = 15
FS.defaults.profile.ba.minTTLET = 20
FS.defaults.profile.hm.minTTLHM = 15

-- initializes FS_TTL - called when the module is loaded
function FS_TTL:OnInitialize()
	FS.modNames = FS.modNames or {}
	FS.modNames["ttl"] = "TimeToLife"

	FS.options.args.ttl = {
		type = "group",
		name = "Time to Life Module",
		order = 9,
		args = {
			enabled = {
				type = "toggle",
				name = "Enabled",
				get = "IsEnabled",
				set = "ToggleEnabled",
				width = "full",
				order = 1,
			},
			shown = {
				type = "toggle",
				name = "Show TTL frame",
				get = "IsTTLshown",
				set = "ToggleTTLshow",
				width = "full",
				order = 2,
			},
			scale = {
				type = "range",
				name = "Frame scale",
				min = 50,
				max = 200,
				bigStep = 10,
				get = "GetScale",
				set = "SetScale",
				order = 3,
			},
			alpha = {
				type = "range",
				name = "Frame alpha",
				min = 10,
				max = 100,
				bigStep = 5,
				get = "GetAlpha",
				set = "SetAlpha",
				order = 4,
			},
		},
	}
	FS.options.args.sps.args.minTTLSpS = {
		type = "range",
		name = "Min. remaining time to live to suggest Serpent Sting",
		min = 0,
		max = 60,
		step = 1,
		width = "double",
		get = "GetMinTTLSpS",
		set = "SetMinTTLSpS",
	}
	FS.options.args.ba.args.minTTLBA = {
		type = "range",
		name = "Min. remaining time to live to suggest Black Arrow",
		min = 0,
		max = 60,
		step = 1,
		width = "double",
		get = "GetMinTTLBA",
		set = "SetMinTTLBA",
	}
	FS.options.args.ba.args.minTTLET = {
		type = "range",
		name = "Min. remaining time to live to suggest Explosive Trap",
		min = 0,
		max = 60,
		step = 1,
		disabled = true,
		width = "double",
		get = "GetMinTTLET",
		set = "SetMinTTLET",
	}
	FS.options.args.hm.args.minTTLHM = {
		type = "range",
		name = "Min. remaining time to live to suggest Hunter's Mark",
		min = 0,
		max = 60,
		step = 1,
		width = "double",
		get = "GetMinTTLHM",
		set = "SetMinTTLHM",
	}
end

-- loads required variables when module is enabled or sets it to disabled if user turned it off
function FS_TTL:OnEnable()
	if FS.db.profile.ttl.enabled then
		self:CheckForDisabled()
		self.tslu = 0
		self.anim = {".  ", ".. ", "..."}
		self.frame = CreateFrame("Frame")
		self.frame:SetScript("OnUpdate", function(this, elapsed) self:OnUpdate(elapsed) end)
		self:GetNumPlayers()
		if FS.db.profile.ttl.shown then	self:CreateTTLFrame() end
	else
		FS:DisableModule("TimeToLife")
	end
end

-- disables the module and deletes all of its existing tables
function FS_TTL:OnDisable()
	self:CheckForDisabled()
	-- if the ttl frame has been enabled this session de-allocate its memory
	if self.frame then
		if self.mov then
			self.mov:Hide()
			self.mov = nil
			self.frm:Hide()
			self.frm = nil
		end
		self:PurgeTTL()
		self.tm = nil
		self.hp = nil
		self.frame:SetScript("OnUpdate", nil)
		self.frame = nil
	end
end

-- basetime for FS_TTL that calls CalcTTL at given intervals
function FS_TTL:OnUpdate(elapsed)
	self.tslu = self.tslu + elapsed
	if self.tslu > FS.db.profile.ttl.updInterval then self:CalcTTL() end
end

-- determine which options need to be activated or deactivated
function FS_TTL:CheckForDisabled()
	if FS.db.profile.ttl.enabled then
		FS.options.args.ttl.args.shown.disabled = false
		FS.options.args.ttl.args.scale.disabled = false
		FS.options.args.ttl.args.alpha.disabled = false
		if FS.db.profile.hm.suggestHM == "never" then FS.options.args.hm.args.minTTLHM.disabled = true
		else FS.options.args.hm.args.minTTLHM.disabled = false
		end
		if FS.db.profile.sps.suggestSpS == "never" then FS.options.args.sps.args.minTTLSpS.disabled = true
		else FS.options.args.sps.args.minTTLSpS.disabled = false
		end
		if FS.db.profile.ba.suggestBA == "never" then FS.options.args.ba.args.minTTLBA.disabled = true
		else FS.options.args.ba.args.minTTLBA.disabled = false
		end
		if FS.db.profile.ba.useTraps == "never" or FS.db.profile.ba.suggestET == "never" then FS.options.args.ba.args.minTTLET.disabled = true
		else FS.options.args.ba.args.minTTLET.disabled = false
		end
	else
		FS.options.args.ttl.args.shown.disabled = true
		FS.options.args.ttl.args.scale.disabled = true
		FS.options.args.ttl.args.alpha.disabled = true
		FS.options.args.sps.args.minTTLSpS.disabled = true
		FS.options.args.ba.args.minTTLBA.disabled = true
		FS.options.args.hm.args.minTTLHM.disabled = true
		FS.options.args.ba.args.minTTLET.disabled = true
	end
end

-- function to unlock the ttl frame if it´s shown
function FS_TTL:UnLock()
	if FS.db.profile.ttl.shown then
		self.mov:SetBackdropColor(0, 0, 0, 0.4)
		self.mov.str:SetTextColor(1, 1, 1, 1)
		self.mov:EnableMouse(true)
	end
end

-- function to lock the ttl frame if it´s shown
function FS_TTL:Lock()
	self.mov:SetBackdropColor(0, 0, 0, 0)
	self.mov.str:SetTextColor(1, 1, 1, 0)
	self.mov:EnableMouse(false)
end

-- called when a valid new target is acquired
function FS_TTL:AcquireTarget()
	if FS.db.profile.ttl.enabled then self:CalcTTL() end
end

-- called when target is lost or not valid
function FS_TTL:LooseTarget()
	if FS.db.profile.ttl.enabled then self:CalcTTL() end
end

-- function to get the Numer of players in party/raid
function FS_TTL:GetNumPlayers()
	FS_TTL.numPlayers = GetNumGroupMembers() + 1
end

-- function to purge ttl tables that of targets that are no longer required
function FS_TTL:PurgeTTL(what)
	local now = GetTime()
	for key in pairs(self.tm) do
		if not what or (what and (now - self.tm[key][1]) > what) then
			FS:RecycleTable(self.tm[key])
			self.tm[key] = nil
			FS:RecycleTable(self.hp[key])
			self.hp[key] = nil
		end
	end
end

-- function to calculate the estimated remaining time to life for current target
function FS_TTL:CalcTTL()
	self.tslu = 0
	local dps, ttl
	local now = now()
	if FS.state ~= "idle" then
		if FS.db.profile.ttl.shown and not self.frm:IsShown() then self.frm:Show() end
		local hp = UnitHealth("target")
		local hpDiff, tmDiff = 0, 0
		self.hp[FS.tarGUID] = self.hp[FS.tarGUID] or FS:GetTable()
		self.tm[FS.tarGUID] = self.tm[FS.tarGUID] or FS:GetTable()
		table.insert(self.hp[FS.tarGUID], 1, hp)
		table.remove(self.hp[FS.tarGUID], FS.db.profile.ttl.numSamples)
		table.insert(self.tm[FS.tarGUID], 1, now)
		table.remove(self.tm[FS.tarGUID], FS.db.profile.ttl.numSamples)
		local lastIndex = table.maxn(self.hp[FS.tarGUID])
		if lastIndex > FS.db.profile.ttl.evalSamples then
			hpDiff = self.hp[FS.tarGUID][lastIndex] - hp
			tDiff = now - self.tm[FS.tarGUID][lastIndex]
			dps = hpDiff / tDiff
			if dps < 0 then dps = 0 end
		elseif self.numPlayers > 1 then
			dps = FS.db.profile.ttl.assumedDps * self.numPlayers * (2 / 3)
		else
			dps = FS.db.profile.ttl.assumedDps
		end
		FS_TTL.delta = hp / (dps + 0.001)
		if dps == 0 then ttl = "∞"
		else ttl = FS_TTL.delta
		end
		if lastIndex <= FS.db.profile.ttl.evalSamples then
			local i = ((lastIndex - 1) % table.maxn(self.anim)) + 1
			dps, ttl = self.anim[i], self.anim[i]
		end
	else
		FS_TTL.delta = nil
		if self.frm:IsShown() then self.frm:Hide() end
		self:PurgeTTL(15)
	end
	if FS.db.profile.ttl.shown then self.frm.str:SetText("DPS: "..FS:Round(dps).." TTL: "..FS:Round(ttl)) end
end

-- function to create the visible ttl frame
function FS_TTL:CreateTTLFrame()
	self.mov = CreateFrame("Frame", "ttlMover", UIParent)
	self.mov:SetFrameStrata("BACKGROUND")
	self.mov:SetWidth(100)
	self.mov:SetHeight(20)
	self.mov:SetScale(FS.db.profile.ttl["scale"]/100)
	self.mov:SetPoint(FS.db.profile.ttl["point"], FS.db.profile.ttl["x"], FS.db.profile.ttl["y"])
	self.mov:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 32,})
	self.mov:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	self.mov:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() FS.db.profile.ttl["point"], _, _, FS.db.profile.ttl["x"], FS.db.profile.ttl["y"] = self:GetPoint() end)
	self.mov:SetMovable(true)
	self.mov:SetClampedToScreen(true)
	self.mov.str = self.mov:CreateFontString("String", "ARTWORK", "GameFontNormal")
	self.mov.str:SetFont(GameFontNormal:GetFont(), 10)
	self.mov.str:SetJustifyH("CENTER")
	self.mov.str:SetJustifyV("TOP")
	self.mov.str:SetPoint("CENTER", self.mov, "CENTER", 0, 20)
	self.mov.str:SetText("TTL Frame")

	-- hide anchor frame if it is locked
	if FS.db.profile.lock then
		self.mov:SetBackdropColor(0, 0, 0, 0)
		self.mov.str:SetTextColor(1, 1, 1, 0)
		self.mov:EnableMouse(false)
	-- otherwise show and make it moveable
	else
		self.mov:SetBackdropColor(0, 0, 0, 0.4)
		self.mov.str:SetTextColor(1, 1, 1, 1)
		self.mov:EnableMouse(true)
	end
	self.frm = CreateFrame("Frame", "ttlFrame", self.mov)
	self.frm:SetWidth(92)
	self.frm:SetHeight(12)
	self.frm:EnableMouse(false)
	self.frm:SetAlpha(FS.db.profile.ttl["alpha"]/100)
	self.frm:SetPoint("CENTER")
	self.frm.str = self.frm:CreateFontString("String", "ARTWORK", "GameFontNormal")
	self.frm.str:SetFont(GameFontNormal:GetFont(), 10)
	self.frm.str:SetJustifyH("CENTER")
	self.frm.str:SetJustifyV("CENTER")
	self.frm.str:SetPoint("CENTER", self.mov, "CENTER", 0, 0)
	self.frm.str:SetTextColor(1, 1, 1, 1)
	self.frm.str:SetText("DPS: - TTL: -")
	self.hp = {}
	self.tm = {}
end

---------------
--- Options ---
---------------

function FS:IsTTLshown(info)
	return FS.db.profile.ttl.shown
end

function FS:ToggleTTLshow(info, value)
	FS.db.profile.ttl.shown = value
	if value then
		self:Print("Time to Life frame is shown now.")
		FS_TTL.frm:Show()
	else
		self:Print("Time to Life frame is no longer shown.")
		FS_TTL.frm:Hide()
	end
end

function FS:GetMinTTLHM(info)
	return FS.db.profile.hm.minTTLHM
end

function FS:SetMinTTLHM(info, value)
	FS.db.profile.hm.minTTLHM = value
	if info.uiType == "cmd" then self:Print("Minimum remaining time to live to suggest Hunter's Mark set to "..value.."s") end
end

function FS:GetMinTTLSpS(info)
	return FS.db.profile.sps.minTTLSpS
end

function FS:SetMinTTLSpS(info, value)
	FS.db.profile.sps.minTTLSpS = value
	if info.uiType == "cmd" then self:Print("Minimum remaining time to live to suggest Serpent Sting set to "..value.."s") end
end

function FS:GetMinTTLBA(info)
	return FS.db.profile.ba.minTTLBA
end

function FS:SetMinTTLBA(info, value)
	FS.db.profile.ba.minTTLBA = value
	if info.uiType == "cmd" then self:Print("Minimum remaining time to live to suggest Black Arrow set to "..value.."s") end
end

function FS:GetMinTTLET(info)
	return FS.db.profile.ba.minTTLET
end

function FS:SetMinTTLET(info, value)
	FS.db.profile.ba.minTTLET = value
	if info.uiType == "cmd" then self:Print("Minimum remaining time to live to suggest Explosive Trap set to "..value.."s") end
end
